/*
 * Copyright 2005 Red Hat, Inc.
 */
#include "../config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <ftw.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <syslog.h>
#include <dbus/dbus.h>
#include "handlers.h"
#include "selinux.h"
#include "util.h"

#define _(_x) _x

#ident "$Id: mkhomedir.c,v 1.1 2005/04/25 18:58:27 nalin Exp $"

static const char *skel;
static struct passwd *pwd;

/* Create a directory with the given permissions, and create leading
 * directories as well. */
static int
my_mkdir(const char *newpath, mode_t mode)
{
	char path[PATH_MAX + 1];
	const char *p;
	int i;
	if (strlen(newpath) < sizeof(path)) {
		/* Walk the path name. */
		memset(path, '\0', sizeof(path));
		for (p = newpath; *p != '\0'; p++) {
			/* If the path so far non-empty, and the next char
			 * would be a directory separator, create the leading
			 * directory.  Accept EEXIST as the only okay error. */
			if ((p > newpath) && (*p == '/')) {
				oddjob_set_selinux_file_creation_context(path,
									 S_IFDIR);
				i = mkdir(path, 0711 | S_IFDIR);
				oddjob_unset_selinux_file_creation_context();
				if ((i == -1) && (errno != EEXIST)) {
					syslog(LOG_ERR, "error creating %s: %m",
					       path);
					return HANDLER_FAILURE;
				}
			}
			path[p - newpath] = *p;
		}
		/* Now create the directory itself. */
		oddjob_set_selinux_file_creation_context(path, mode | S_IFDIR);
		i = mkdir(path, mode);
		oddjob_unset_selinux_file_creation_context();
		if (i == -1) {
			syslog(LOG_ERR, "error creating %s: %m", path);
			return HANDLER_FAILURE;
		}
		return 0;
	} else {
		errno = EINVAL;
		syslog(LOG_ERR, "pathname (%s) is too long", newpath);
		return HANDLER_INVALID_INVOCATION;
	}
}
/* Given the path of an item somewhere in the skeleton directory, create as
 * identical as possible a copy in the destination tree. */
static int
copy_single_item(const char *source, const struct stat *sb,
		 int flag, struct FTW *s)
{
	int sfd, dfd, i;
	char target[PATH_MAX + 1], newpath[PATH_MAX + 1];
	unsigned char buf[BUFSIZ];
	/* Generate the name of the new item. */
	if (snprintf(newpath, sizeof(newpath), "%s%s",
		     pwd->pw_dir,
		     source + strlen(skel)) > (int) sizeof(newpath)) {
		/* The path would be too long(!), so give up. */
		syslog(LOG_ERR, "pathname (%s%s) would be too long",
		       pwd->pw_dir, source + strlen(skel));
		return HANDLER_INVALID_INVOCATION;
	}
	switch (flag) {
	case FTW_SL:
		/* It's a symlink.  Read its target and create a copy. */
		memset(&target, '\0', sizeof(target));
		if (readlink(source, target, sizeof(target) - 1)) {
			oddjob_set_selinux_file_creation_context(newpath,
								 sb->st_mode |
								 S_IFLNK);
			if (symlink(target, newpath) == 0) {
				if (lchown(newpath, pwd->pw_uid, pwd->pw_gid) == -1) {
					unlink(newpath);
				}
			} else {
				oddjob_unset_selinux_file_creation_context();
				syslog(LOG_ERR, "error creating %s: %m",
				       target);
				return HANDLER_FAILURE;
			}
			oddjob_unset_selinux_file_creation_context();
			return 0;
		}
		break;
	case FTW_F:
		/* It's a file.  Make a copy. */
		sfd = open(source, O_RDONLY);
		if (sfd != -1) {
			oddjob_set_selinux_file_creation_context(newpath,
								 sb->st_mode |
								 S_IFREG);
			dfd = open(newpath, O_WRONLY | O_CREAT | O_EXCL,
				   sb->st_mode);
			if (dfd != -1) {
				while ((i = read(sfd, buf, sizeof(buf))) > 0) {
					retry_write(dfd, buf, i);
				}
				if (fchown(dfd, pwd->pw_uid, pwd->pw_gid) == -1) {
					unlink(newpath);
				} else {
					if (fchmod(dfd, sb->st_mode) == -1) {
						unlink(newpath);
					}
				}
				close(dfd);
			}
			close(sfd);
			oddjob_unset_selinux_file_creation_context();
		} else {
			syslog(LOG_ERR, "error opening %s: %m", source);
			return HANDLER_FAILURE;
		}
		return 0;
	case FTW_D:
		/* It's a directory.  Make one with the same name and
		 * permissions, but owned by the target user. */
		if (my_mkdir(newpath, sb->st_mode) == 0) {
			if (chown(newpath, pwd->pw_uid, pwd->pw_gid) == -1) {
				rmdir(newpath);
			}
		} else {
			if (errno != EEXIST) {
				return HANDLER_FAILURE;
			}
		}
		return 0;
	case FTW_NS:
	default:
		return 0;
	}
	return 0;
}

/*
 * get_skel_dir
 *
 * Returns: the location on the filesystem where the contents of a new user's
 * home directory should be found.  FIXME: consult /etc/default/useradd.
 *
 */
static const char *
get_skel_dir(void)
{
	return "/etc/skel";
}

/* Create a copy of /etc/skel in the named user's home directory. */
static int
mkhomedir(const char *user)
{
	struct stat st;

	/* Now make sure that the user
	   a) exists
	   b) has a home directory specified which is
	      1) an absolute pathname
	      2) not an empty string
	      3) not already there */
	pwd = getpwnam(user);
	if (pwd == NULL) {
		syslog(LOG_ERR, "could not look up location of home directory "
		       "for %s", user);
		return HANDLER_INVALID_INVOCATION;
	}
	if (pwd->pw_dir == NULL) {
		syslog(LOG_ERR, "user %s has NULL home directory", user);
		return HANDLER_INVALID_INVOCATION;
	}
	if ((strlen(pwd->pw_dir) == 0) || (pwd->pw_dir[0] != '/')) {
		syslog(LOG_ERR, "user %s has weird home directory (%s)", user,
		       pwd->pw_dir);
		return HANDLER_INVALID_INVOCATION;
	}
	if ((lstat(pwd->pw_dir, &st) == -1) && (errno == ENOENT)) {
		/* Figure out which location we're using as a
		 * template. */
		skel = get_skel_dir();
		if (skel != NULL) {
			/* Set the text of the result message. */
			printf(_("Creating home directory for %s."), user);
			/* Walk the template tree and make a copy. */
			return nftw(get_skel_dir(), copy_single_item, 5,
				    FTW_PHYS);
		}
	}
	return 0;
}

int
main(int argc, char **argv)
{
	char arg[LINE_MAX];
	int i;
	openlog("mkhomedir", LOG_PID, LOG_DAEMON);
	if (fgets(arg, sizeof(arg), stdin) != NULL) {
		i = strcspn(arg, "\r\n");
		arg[i] = '\0';
		if (strlen(arg) > 0) {
			return mkhomedir(arg);
		}
	}
	closelog();
	return HANDLER_INVALID_INVOCATION;
}
