/*
 * Copyright 2005 Red Hat, Inc.
 */
#include "../config.h"
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <getopt.h>
#include <glob.h>
#include <grp.h>
#include <pwd.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dbus/dbus.h>
#include <libxml/xmlreader.h>
#include "common.h"
#include "handlers.h"
#include "oddjob_dbus.h"
#include "util.h"

#ident "$Id: oddjobd.c,v 1.2 2005/04/25 20:01:28 nalin Exp $"

static struct {
	struct method {
		/* method name */
		char *method;
		/* method type -- internal function, or call a helper */
		enum method_type {
			method_internal,
			method_external,
		} type;
		int n_arguments;
		/* function to call, only applies to internal methods */
		oddjob_dbus_handler *handler;
		void *handler_data;
		/* path to exec and whether or not we prepend the caller's name
		 * to the list of arguments, only applies to external methods */
		char *execpath;
		dbus_bool_t prepend_user;
		/* ACLs for the method */
		struct method_acl {
			enum method_acl_sense {
				method_acl_allow,
				method_acl_deny
			} sense;
			dbus_bool_t apply_user;
			char *user;
			dbus_bool_t apply_min_uid;
			unsigned long min_uid;
			dbus_bool_t apply_max_uid;
			unsigned long max_uid;
			struct method_acl *next;
		} *acl;
	} *methods;
	int n_methods;
	/* flags */
	int debug;
	int quit;
} globals = {
	.methods = NULL,
	.n_methods = 0,
	.debug = 0,
	.quit = 0,
};

static void oddjobd_exec_method(struct oddjob_dbus_context *ctx,
				struct oddjob_dbus_message *msg,
				const char *user,
				unsigned long uid,
				void *data);
static dbus_bool_t load_config(const char *file, dbus_bool_t ignore_missing);

/* Check if a single ACL entry matches the given user and UID. */
static dbus_bool_t
check_ac_match(struct method_acl *acl, const char *user, unsigned long uid)
{
	int fnflags = FNM_NOESCAPE;
	return ((!acl->apply_user) ||
		(fnmatch(acl->user, user, fnflags) == 0)) &&
	       ((!acl->apply_min_uid) || (uid >= acl->min_uid)) &&
	       ((!acl->apply_max_uid) || (uid <= acl->max_uid));
}

/* Check if there is a match for the given user and UID in the given ACL,
   returning TRUE if there's an access rule which allows the user acces. */
static dbus_bool_t
check_acl(struct method_acl *acl, const char *user, unsigned long uid)
{
	struct method_acl *i;
	for (i = acl; i != NULL; i = i->next) {
		if (i->sense == method_acl_allow) {
			if (check_ac_match(i, user, uid)) {
				if (globals.debug) {
					fprintf(stderr, "Matched allow rule (");
					if (i->apply_user) {
						fprintf(stderr, " user=%s ",
							i->user);
					}
					if (i->apply_min_uid) {
						fprintf(stderr, " min_uid=%lu ",
							i->min_uid);
					}
					if (i->apply_max_uid) {
						fprintf(stderr, " max_uid=%lu ",
							i->max_uid);
					}
					fprintf(stderr, ")\n");
				}
				return TRUE;
			}
		}
	}
	for (i = acl; i != NULL; i = i->next) {
		if (i->sense == method_acl_deny) {
			if (check_ac_match(i, user, uid)) {
				if (globals.debug) {
					fprintf(stderr, "Matched deny rule (");
					if (i->apply_user) {
						fprintf(stderr, " user=%s ",
							i->user);
					}
					if (i->apply_min_uid) {
						fprintf(stderr, " min_uid=%lu ",
							i->min_uid);
					}
					if (i->apply_max_uid) {
						fprintf(stderr, " max_uid=%lu ",
							i->max_uid);
					}
					fprintf(stderr, ")\n");
				}
				return FALSE;
			}
		}
	}
	if (globals.debug) {
		fprintf(stderr, "Fell through ACL, denying access.\n");
	}
	return FALSE;
}
static dbus_bool_t
check_named_acl(const char *method, const char *user, unsigned long uid)
{
	int i;
	for (i = 0; i < globals.n_methods; i++) {
		if (strcmp(globals.methods[i].method, method) == 0) {
			return check_acl(globals.methods[i].acl, user, uid);
		}
	}
	if (globals.debug) {
		fprintf(stderr, "Attempted to check ACL for unknown method, "
			"denying access.\n");
	}
	return FALSE;
}

/* Load ACLs from the configuration file. */
static dbus_bool_t
load_config_xml_node_name_is(xmlNodePtr node, const char *name)
{
	return (node->name != NULL) &&
	       (xmlStrcmp(node->name, (xmlChar*) name) == 0);
}
static dbus_bool_t
load_config_xml_attr_name_is(xmlAttrPtr attr, const char *name)
{
	return (attr->name != NULL) &&
	       (xmlStrcmp(attr->name, (xmlChar*) name) == 0);
}
static const char *
load_config_xml_attr_data(xmlAttrPtr attr)
{
	xmlNodePtr child;
	for (child = attr->children; child != NULL; child = child->next) {
		if ((child->type == XML_TEXT_NODE) &&
		    (child->content != NULL)) {
			return (const char *) child->content;
		}
	}
	return "";
}
static const char *
load_config_xml_node_data(xmlNodePtr node)
{
	xmlNodePtr child;
	for (child = node->children; child != NULL; child = child->next) {
		if ((child->type == XML_TEXT_NODE) &&
		    (child->content != NULL)) {
			return (const char *) child->content;
		}
	}
	return "";
}
static dbus_bool_t
load_config_oddjobconfig_oddjob_access(xmlNodePtr cur,
				       const char *method,
				       enum method_acl_sense sense)
{
	xmlAttrPtr attr;
	struct method_acl *ac, *acl;
	int i;
	const char *p;
	char *q;

	ac = oddjob_malloc(sizeof(*ac));
	if (ac == NULL) {
		fprintf(stderr, "Out of memory\n");
		return FALSE;
	}
	memset(ac, 0, sizeof(*ac));

	ac->sense = sense;
	if (globals.debug) {
		fprintf(stderr, "Add new %s%s rule ( ",
			sense == method_acl_allow ? "allow" : "",
			sense == method_acl_deny ? "deny" : "");
	}
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "user")) {
			ac->apply_user = TRUE;
			ac->user = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "user=\"%s\" ", ac->user);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "min_uid")) {
			p = load_config_xml_attr_data(attr);
			ac->min_uid = strtoul(p, &q, 0);
			if ((*p != '\0') && (q != NULL) && (*q == '\0')) {
				ac->apply_min_uid = TRUE;
				if (globals.debug) {
					fprintf(stderr, "min_uid=\"%lu\" ",
						ac->min_uid);
				}
			}
		} else
		if (load_config_xml_attr_name_is(attr, "max_uid")) {
			p = load_config_xml_attr_data(attr);
			ac->max_uid = strtoul(p, &q, 0);
			if ((*p != '\0') && (q != NULL) && (*q == '\0')) {
				ac->apply_max_uid = TRUE;
				if (globals.debug) {
					fprintf(stderr, "max_uid=\"%lu\" ",
						ac->max_uid);
				}
			}
		}
	}
	if (globals.debug) {
		fprintf(stderr, ") for method \"%s\".\n", method);
	}
	for (i = 0; i < globals.n_methods; i++) {
		if (strcmp(method, globals.methods[i].method) == 0) {
			if (globals.methods[i].acl == NULL) {
				globals.methods[i].acl = ac;
			} else {
				acl = globals.methods[i].acl;
				while (acl->next != NULL) {
					acl = acl->next;
				}
				acl->next = ac;
			}
			return TRUE;
		}
	}
	fprintf(stderr, "Unknown method \"%s\"\n", method);
	free(ac->user);
	free(ac);
	return FALSE;
}
static dbus_bool_t
load_config_oddjobconfig_include(xmlNodePtr cur)
{
	xmlAttrPtr attr;
	dbus_bool_t ignore_missing, okay;
	glob_t globbed;
	unsigned int i;

	ignore_missing = FALSE;
	if (globals.debug) {
		fprintf(stderr, "Include(");
	}
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "ignore_missing")) {
			if (globals.debug) {
				fprintf(stderr, "ignore_missing=\"%s\",",
					load_config_xml_attr_data(attr));
			}
			if (strcmp(load_config_xml_attr_data(attr),
				   "yes") == 0) {
				ignore_missing = TRUE;
			}
		}
	}
	if (globals.debug) {
		fprintf(stderr, "\"%s\").\n", load_config_xml_node_data(cur));
	}

	memset(&globbed, 0, sizeof(globbed));
	if (glob(load_config_xml_node_data(cur), GLOB_NOCHECK,
		 NULL, &globbed) != 0) {
		if (globals.debug) {
			fprintf(stderr, "glob: %s: %s.\n",
				load_config_xml_node_data(cur),
				strerror(errno));
		}
		return FALSE;
	}

	for (i = 0, okay = TRUE; i < globbed.gl_pathc; i++) {
		okay = load_config(globbed.gl_pathv[i],
				   ignore_missing) && okay;
	}

	globfree(&globbed);

	return okay;
}

static void
method_add(const char *method, enum method_type type,
	   oddjob_dbus_handler handler,
	   void *handler_data,
	   int n_arguments, dbus_bool_t prepend_user)
{
	oddjob_resize_array((void**)&globals.methods,
			    sizeof(globals.methods[0]),
			    globals.n_methods,
			    globals.n_methods + 1);
	globals.methods[globals.n_methods].method = oddjob_strdup(method);
	globals.methods[globals.n_methods].type = type;
	globals.methods[globals.n_methods].handler = handler;
	globals.methods[globals.n_methods].handler_data = handler_data;
	globals.methods[globals.n_methods].n_arguments = n_arguments;
	globals.methods[globals.n_methods].prepend_user = prepend_user;
	globals.n_methods++;
}

static dbus_bool_t
load_config_oddjobconfig_oddjob_helper(xmlNodePtr cur, const char *method)
{
	xmlAttrPtr attr;
	int i, n_arguments;
	const char *execpath, *arguments, *prepend;
	char *p;
	dbus_bool_t prepend_user;
	for (i = 0; i < globals.n_methods; i++) {
		if (strcmp(globals.methods[i].method, method) == 0) {
			break;
		}
	}
	if (i >= globals.n_methods) {
		fprintf(stderr, "internal error %d (no method named \"%s\")!",
			__LINE__, method);
		return FALSE;
	}
	if (globals.methods[i].type != method_external) {
		fprintf(stderr, "<helper> node for internal method \"%s\" is "
			"not allowed!\n", method);
		return FALSE;
	}
	execpath = NULL;
	arguments = NULL;
	prepend = NULL;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "exec")) {
			execpath = load_config_xml_attr_data(attr);
		}
		if (load_config_xml_attr_name_is(attr, "arguments")) {
			arguments = load_config_xml_attr_data(attr);
		}
		if (load_config_xml_attr_name_is(attr, "prepend_user_name")) {
			prepend = load_config_xml_attr_data(attr);
		}
	}
	if (execpath == NULL) {
		fprintf(stderr, "Required attribute \"exec\" not given!\n");
		return FALSE;
	}
	if (execpath[0] != '/') {
		fprintf(stderr, "\"exec\" value \"%s\" is not "
			"an absolute path!\n", execpath);
		return FALSE;
	}
	n_arguments = 0;
	if (arguments != NULL) {
		n_arguments = strtol(arguments, &p, 0);
		if ((*arguments == '\0') || (p == NULL) || (*p != '\0')) {
			fprintf(stderr, "Invalid argument count \"%s\"!\n",
				arguments);
			return FALSE;
		}
	}
	prepend_user = FALSE;
	if (prepend != NULL) {
		prepend_user = (strcmp(prepend, "yes") == 0);
	}
	globals.methods[i].execpath = oddjob_strdup(execpath);
	globals.methods[i].n_arguments = n_arguments;
	globals.methods[i].prepend_user = prepend_user;
	if (globals.debug) {
		fprintf(stderr, "Configure external method \"%s\", exec=\"%s\","
			" arguments=%d, prepend_user_name=%s.\n",
			method, execpath, n_arguments,
			prepend_user ? "yes" : "no");
	}
	return TRUE;
}

static dbus_bool_t
load_config_oddjobconfig_oddjob(xmlNodePtr cur)
{
	xmlAttrPtr attr;
	xmlNodePtr child;
	const char *method;
	dbus_bool_t added_here;
	int i;

	method = NULL;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "method")) {
			method = load_config_xml_attr_data(attr);
		}
	}
	if (method == NULL) {
		fprintf(stderr, "<oddjob> node has no \"method\" attribute!\n");
		return FALSE;
	}
	added_here = FALSE;
	for (i = 0; i < globals.n_methods; i++) {
		if (strcmp(method, globals.methods[i].method) == 0) {
			break;
		}
	}
	if (i >= globals.n_methods) {
		/* check if it's an external method */
		for (child = cur->children;
		     child != NULL;
		     child = child->next) {
			if (load_config_xml_node_name_is(child, "helper")) {
				/* We'd pass the address of the method here,
				 * but the array may be reallocated, so let
				 * main() set it up. */
				method_add(method, method_external,
					   oddjobd_exec_method, NULL,
					   0, FALSE);
				load_config_oddjobconfig_oddjob_helper(child,
								       method);
				break;
			}
		}
		added_here = TRUE;
	}
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "helper")) {
			if (added_here == FALSE) {
				fprintf(stderr, "duplicate <helper> node for "
					"\"%s\" method!\n", method);
				return FALSE;
			}
		}
		if (load_config_xml_node_name_is(child, "deny")) {
			load_config_oddjobconfig_oddjob_access(child,
							       method,
							       method_acl_deny);
		}
		if (load_config_xml_node_name_is(child, "allow")) {
			load_config_oddjobconfig_oddjob_access(child,
							       method,
							       method_acl_allow);
		}
	}

	return TRUE;
}
static dbus_bool_t
load_config_oddjobconfig(xmlNodePtr cur)
{
	xmlNodePtr child;
	dbus_bool_t parsed;
	parsed = TRUE;
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "include")) {
			parsed = parsed &&
				 load_config_oddjobconfig_include(child);
		}
		if (load_config_xml_node_name_is(child, "oddjob")) {
			parsed = parsed &&
				 load_config_oddjobconfig_oddjob(child);
		}
	}
	return parsed;
}
static dbus_bool_t
load_config(const char *filename, dbus_bool_t ignore_missing)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	struct stat st;
	dbus_bool_t parsed;

	if (filename == NULL) {
		filename = SYSCONFDIR "/" PACKAGE "d.conf";
	}

	if (ignore_missing &&
	    (stat(filename, &st) == -1) && (errno == ENOENT)) {
		if (globals.debug) {
			fprintf(stderr, "Ignoring missing file \"%s\".\n",
				filename);
		}
		return TRUE;
	}

	if (globals.debug) {
		fprintf(stderr, "Parsing configuration file \"%s\".\n",
			filename);
	}
	doc = xmlParseFile(filename);
	if (doc == NULL) {
		fprintf(stderr, "Error parsing configuration from \"%s\".\n",
			filename);
		return FALSE;
	}

	parsed = FALSE;
	cur = xmlDocGetRootElement(doc);
	if (cur != NULL) {
		if (load_config_xml_node_name_is(cur, "oddjobconfig")) {
			parsed = load_config_oddjobconfig(cur);
		}
	}

	xmlFreeDoc(doc);

	return parsed;
}
static void
unload_config(void)
{
	struct method_acl *acl, *tofree;
	int i;

	for (i = 0; i < globals.n_methods; i++) {
		oddjob_free(globals.methods[i].method);
		globals.methods[i].method = NULL;
		if (globals.methods[i].execpath) {
			oddjob_free(globals.methods[i].execpath);
			globals.methods[i].execpath = NULL;
		}
		acl = globals.methods[i].acl;
		while (acl != NULL) {
			tofree = acl;
			acl = acl->next;
			oddjob_free(tofree->user);
			oddjob_free(tofree);
		}
		globals.methods[i].acl = NULL;
	}
	oddjob_free(globals.methods);
	globals.methods = NULL;
	globals.n_methods = 0;
}

static void
oddjobd_exec_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	struct method *method;
	int exec_status[2], child_in[2], child_out[2], null[3], i;
	unsigned int n;
	char outbuf[BUFSIZ], *output;
	pid_t child;
	unsigned char exec_errno, output_char;
	int status;
	const char *arg, *helper_name;

	method = data;

	if (globals.debug) {
		fprintf(stderr, "Received request for \"%s\" from %s.\n",
			method->method, user);
	}
	if (!check_acl(method->acl, user, uid)) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"%s\" denied to %s.\n",
				method->method, user);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}

	/* Open /dev/null three times, just in case stdio aren't already set. */
	if (globals.debug) {
		fprintf(stderr, "Opening /dev/null 3 times.\n");
	}
	for (n = 0; n < sizeof(null) / sizeof(null[0]); n++) {
		null[n] = open("/dev/null", O_RDWR);
		if (null[n] == -1) {
			snprintf(outbuf, sizeof(outbuf),
				"Error at open(/dev/null) %d: %s.",
				n + 1, strerror(errno));
			if (globals.debug) {
				fprintf(stderr, "%s\n", outbuf);
			}
			oddjob_dbus_send_message_response_error(msg,
								ODDJOB_INTERFACE_NAME
								".error.internal",
								outbuf);
			/* Clean up any /dev/null descriptors which we've
			 * already opened. */
			i = n - 1;
			while (i >= 0) {
				close(null[i]);
				i--;
			}
			return;
		}
	}
	/* Create a pipe for returning the exec() failure code. */
	if (pipe(exec_status) == -1) {
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(outbuf, sizeof(outbuf),
			"Error at pipe(1): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							outbuf);
		return;
	}
	/* Create a pair of pipes for stdio. */
	if (pipe(child_in) == -1) {
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(outbuf, sizeof(outbuf),
			"Error at pipe(2): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							outbuf);
		return;
	}
	if (pipe(child_out) == -1) {
		close(child_in[0]);
		close(child_in[1]);
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(outbuf, sizeof(outbuf),
			"Error at pipe(3): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							outbuf);
		return;
	}
	/* At this point we *know* that none of p or s are on stdio, so we no
	 * longer need null[]. */
	close(null[0]);
	close(null[1]);
	close(null[2]);
	/* Here's where the heavy lifting starts. */
	child = fork();
	switch (child) {
	case -1:
		close(exec_status[0]);
		close(exec_status[1]);
		close(child_in[0]);
		close(child_in[1]);
		close(child_out[0]);
		close(child_out[1]);
		snprintf(outbuf, sizeof(outbuf),
			 "Error at fork(): %s.\n", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							outbuf);
		return;
		break;
	case 0:
		if (globals.debug) {
			fprintf(stderr, "Child preparing to exec %s.\n",
				method->execpath);
		}
		/* we're the child -- close the read end of the error pipe, and
		 * mark the write end to be closed on successful exec() */
		close(exec_status[0]);
		fcntl(exec_status[1], F_SETFD, FD_CLOEXEC);
		/* set up stdio */
		close(child_in[1]);
		close(child_out[0]);
		dup2(child_in[0], 0);
		dup2(child_out[1], 1);
		dup2(child_out[1], 2);
		/* close any open descriptors which aren't the exec status pipe
		 * or stdio */
		for (i = sysconf(_SC_OPEN_MAX) - 1; i > 2; i--) {
			if (i != exec_status[1]) {
				close(i);
			}
		}
		/* run the helper */
		helper_name = method->execpath;
		if (strchr(helper_name, '/') != NULL) {
			helper_name = strrchr(helper_name, '/') + 1;
		}
		execl(method->execpath, helper_name, NULL);
		/* uh-oh. send errno to the caller and bail */
		exec_errno = errno;
		write(exec_status[1], &exec_errno, 1);
		_exit(-1);
		break;
	default:
		break;
	}
	/* check the exec() status */
	close(exec_status[1]);
	close(child_in[0]);
	close(child_out[1]);
	while ((i = (read(exec_status[0], &exec_errno, 1))) == -1) {
		continue;
	}
	close(exec_status[0]);
	/* if we got a byte, then it's an exec() error */
	if (i == 1) {
		close(child_in[1]);
		close(child_out[0]);
		snprintf(outbuf, sizeof(outbuf),
			 "Child signalled exec() error: %s.",
			 strerror(exec_errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							outbuf);
		return;
	}
	/* exec() succeeded -- send arguments */
	output = outbuf;
	if (method->prepend_user) {
		if (globals.debug) {
			fprintf(stderr, "Sending user \"%s\".\n", user);
		}
		write(child_in[1], user, strlen(user));
		write(child_in[1], "\n", 1);
	}
	for (i = 0; i < method->n_arguments; i++) {
		arg = oddjob_dbus_message_get_arg(msg, i);
		if (globals.debug) {
			fprintf(stderr, "Sending argument \"%s\".\n", arg);
		}
		write(child_in[1], arg, strlen(arg));
		write(child_in[1], "\n", 1);
	}
	close(child_in[1]);
	/* read back any text which the helper sends to us */
	memset(outbuf, '\0', sizeof(outbuf));
	while ((output < outbuf + sizeof(outbuf)) &&
	       (i = (read(child_out[0], &output_char, 1) != 0))) {
		if (i == 1) {
			*output++ = output_char;
		}
		if (i == -1) {
			if ((errno == EAGAIN) || (errno == EINTR)) {
				continue;
			} else {
				break;
			}
		}
	}
	close(child_out[0]);
	/* check the helper's return code */
	if (waitpid(child, &status, 0) == -1) {
		snprintf(outbuf, sizeof(outbuf),
			 "Error at waitpid(): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							outbuf);
		return;
	}
	if (WIFSIGNALED(status)) {
		snprintf(outbuf, sizeof(outbuf),
			 "Child exited on signal %d.",
			 WTERMSIG(status));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							outbuf);
	}
	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) {
		if (globals.debug) {
			fprintf(stderr, "Child completed its task.\n");
		}
		oddjob_dbus_send_message_response_success(msg, outbuf);
	}
	if (WIFEXITED(status)) {
		i = WEXITSTATUS(status);
		snprintf(outbuf, sizeof(outbuf),
			 "helper exited with error %d", i);
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							outbuf);
	}
}

static void
oddjobd_list_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	int i;
	char *result, *t;
	if (!check_named_acl("list", user, uid)) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"list\" denied.\n");
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}
	result = NULL;
	for (i = 0; i < globals.n_methods; i++) {
		if (check_acl(globals.methods[i].acl, user, uid)) {
			if (result == NULL) {
				result = oddjob_strdup(globals.methods[i].method);
			} else {
				t = oddjob_malloc0(strlen(result) + 1 +
						   strlen(globals.methods[i].method) + 1);
				sprintf(t, "%s,%s", result,
					globals.methods[i].method);
				oddjob_free(result);
				result = t;
			}
		}
	}
	oddjob_dbus_send_message_response_success(msg, result);
	oddjob_free(result);
}

static void
oddjobd_quit_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	if (!check_named_acl("quit", user, uid)) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"%s\" denied to %s.\n",
				"quit", user);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}
	if (globals.debug) {
		fprintf(stderr, "Shutting down.\n");
	}
	oddjob_dbus_send_message_response_success(msg, NULL);
	globals.quit++;
}

int
main(int argc, char **argv)
{
	struct oddjob_dbus_context *ctx;
	FILE *pidfile;
	gid_t gid;
	int c, i;
	const char *basename, *usage;
	struct {
		dbus_bool_t nofork;
		const char *pidfile;
		const char *config;
	} options = {
		.nofork = FALSE,
		.pidfile = NULL,
		.config = NULL,
	};

	/* If we need to provide a usage message, then we need the basename of
	 * the binary. */
	basename = argv[0];
	if (strchr(basename, '/')) {
		basename = strrchr(basename, '/') + 1;
	}

	/* Process command-line options. */
	usage = "Usage: %s [-d] [-n] [-p pidfile] [-c configfile]\n";
	while ((c = getopt(argc, argv, "dnp:c:")) != -1) {
		switch (c) {
		case 'd':
			globals.debug++;
			/* fall through */
		case 'n':
			options.nofork = TRUE;
			break;
		case 'p':
			options.pidfile = optarg;
			break;
		case 'c':
			options.config = optarg;
			break;
		default:
			printf(usage, basename);
			return 1;
			break;
		}
	}
	if ((argc - optind) > 0) {
		printf(usage, basename);
		return 1;
	}

	/* Add our built-in handlers to our method table. */
	method_add("quit", method_internal,
		   oddjobd_quit_method, NULL,
		   0, FALSE);
	method_add("list", method_internal,
		   oddjobd_list_method, NULL,
		   0, FALSE);

	/* Read in ACLs and definitions for external handlers. */
	if (!load_config(options.config, FALSE)) {
		fprintf(stderr, "Error loading configuration!\n");
		return 1;
	}

	/* Open a connection to the system message bus. */
	ctx = oddjob_dbus_listener_new(ODDJOB_SERVICE_NAME);
	if (ctx == NULL) {
		fprintf(stderr, "Error connecting to system D-BUS!\n");
		return 2;
	}

	/* Now add our handlers to the list which our D-BUS layer knows. */
	for (i = 0; i < globals.n_methods; i++) {
		/* If the handler used NULL, give it the method table address,
		 * which wasn't fixed until now. */
		oddjob_dbus_listener_add_method(ctx,
						ODDJOB_OBJECT_PATH,
						ODDJOB_INTERFACE_NAME,
						globals.methods[i].method,
						globals.methods[i].n_arguments,
						globals.methods[i].handler,
						globals.methods[i].handler_data ?
						globals.methods[i].handler_data :
						&globals.methods[i]);
	}

	/* Clear the supplemental group list.  We assume that we were started
	 * with the right UID/GID values, but there should be no need to keep
	 * the rest of the group memberships? */
	gid = getgid();
	if (setgroups(1, &gid) == -1) {
		fprintf(stderr, "Error clearing supplemental group list!\n");
		return 1;
	}

	/* Change the working directory to the root directory. */
	if (chdir("/") == -1) {
		fprintf(stderr, "Error changing working directory to \"/\"!\n");
		return 1;
	}

	/* Create a pidfile. */
	if (options.pidfile != NULL) {
		pidfile = fopen(options.pidfile, "w");
		if (pidfile == NULL) {
			fprintf(stderr, "Error creating pidfile!\n");
			return 1;
		}
		fprintf(pidfile, "%lu", (long) getpid());
		fclose(pidfile);
	}

	/* Become a daemon, changing the working directory to the root,
	 * dissociating from the controlling terminal (if there is one),
	 * calling fork(), and letting the parent exit. */
	if (!options.nofork) {
		if (daemon(0, 0) != 0) {
			fprintf(stderr, "Error becoming a daemon!\n");
			if (options.pidfile != NULL) {
				unlink(options.pidfile);
			}
			return 1;
		}
	}

	/* Now sit! */
	oddjob_dbus_main_init(ctx);
	globals.quit = 0;
	while (globals.quit == 0) {
		oddjob_dbus_main_iterate(ctx);
	}
	oddjob_dbus_main_done(ctx);

	/* Clean up and exit. */
	unload_config();
	oddjob_dbus_listener_free(ctx);
	if (options.pidfile != NULL) {
		unlink(options.pidfile);
	}

	return 0;
}
